# psim を利用するためのおまじない -- ここから
import winreg, sys, os
reg = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Wow6432Node\\Mathematical Systems Inc.\\S-Quattro Simulation System\\INFO")
rootdir = winreg.QueryValueEx(reg, "ROOTDIR")[0]
os.environ["PATH"] = rootdir + ";" + os.environ["PATH"]
sys.path.insert(0, os.path.join(rootdir, "lib3"))
# psim を利用するためのおまじない -- ここまで

from psim import *
import matplotlib.pyplot as plt
import matplotlib
import pandas as pd
import pdb

# 工程定義
mlist = pd.read_csv("mlist.csv", encoding = "cp932")
mlist = mlist.values.tolist()

# 製品クラス
class Item:
    def __init__(self, name):
        self.name = name
        self.dic = {}
    def recordEnter(self, mname, time):
        self.dic[mname] = [time, time]
    def recordExit(self, mname, time):
        self.dic[mname][1] = time
    def getLog(self):
        return [(m, self.name, t0, t1) for m, (t0, t1) in self.dic.items()]
    def __repr__(self):
        return self.name

# 加工機械
class Machine:
    def __init__(self, name, nbuf, npar, mu, sig):
        self.name = name
        self.nbuf = nbuf
        self.npar = npar
        self.mu = mu
        self.sig = sig
        if self.nbuf == 0:
            self.inputQueue = Store("empty")
        else:
            self.inputQueue = Store(self.nbuf)
        for i in range(self.npar):
            activate(self.proc)(i)
        self.next = None
    def proc(self, i):
        def machinename(i):
            if self.npar == 1:
                return self.name
            else:
                return "%s-%d" % (self.name, i + 1)
        while True:
            result = yield self.inputQueue.get1(name = "item")
            item = result["item"]
            item.recordEnter(machinename(i), now())
            yield pause(np.random.normal(self.mu, self.sig))
            item.recordExit(machinename(i), now())
            if self.next is not None:
                yield self.next.put(item)
    def setNext(self, machine):
        self.next = machine
        return self.name
    def put(self, item):
        def _():
            #print(f"putting {item} into {self.name} at {now()}")
            yield self.inputQueue.put1(item)
            #print(f"put {item} into {self.name} at {now()}")
        return call(_)()

# 投入数
nitems = 30

initialize()

# 加工機械を作成
machines = []    
for i, args in enumerate(mlist):
    m = Machine(*args)
    machines.append(m)

# 加工機械を結合
for i in range(len(machines) - 1):
    machines[i].setNext(machines[i + 1])
    
items = []

def put(item):
    yield machines[0].put(item)
    
# 生産ラインに製品を投入するプロセス
def proc():
    for i in range(nitems):
        item = Item("item%2d" % i)
        items.append(item)
        activate(put)(item)
    yield alwaysTrue()

# シミュレーション開始
activate(proc)()        
start(until = None)

# ガントチャート作成
def plot_gantt(ax, sched, title):
    machines = set([machine for machine, job, start, end in sched])
    machines = sorted(list(machines))
    machines = {m: i for i, m in enumerate(machines)}
    items = set([job for machine, job, start, end in sched])
    items = sorted(list(items))
    items = {m: i for i, m in enumerate(items)}
    starttimes = [start for machine, job, start, end in sched]
    endtimes = [end for machine, job, start, end in sched]
    colors = matplotlib.cm.tab20.colors
    for machine, job, start, end in sched:
        ax.broken_barh([(start, end - start)], (machines[machine] - 0.4, 0.8),
                       color = colors[items[job] % len(colors)])
    ax.set_yticks(range(len(machines)))
    ax.set_yticklabels(machines)
    ax.set_xlabel("time")
    ax.set_xlim((min(starttimes), max(endtimes)))
    ax.invert_yaxis()
    ax.set_title(title)

sched = []
for item in items:
    sched.extend(item.getLog())

endtimes = [end for machine, job, start, end in sched]
print(max(endtimes)) # 終了時間

fig, axes = plt.subplots(1, 1)
plot_gantt(axes, sched, "gantt chart")
plt.show()
#plt.savefig("gantt.png")

