# coding: utf-8

# psim を利用するためのおまじない -- ここから
import winreg, sys, os
reg = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Wow6432Node\\Mathematical Systems Inc.\\S-Quattro Simulation System\\INFO")
rootdir = winreg.QueryValueEx(reg, "ROOTDIR")[0]
os.environ["PATH"] = rootdir + ";" + os.environ["PATH"]
sys.path.insert(0, os.path.join(rootdir, "lib3"))
# psim を利用するためのおまじない -- ここまで

import xlwings as xw

from psim import *

def runSimulation():
    initialize()
    
    lmd = xw.Range("B3").value # 平均到着率
    mu = xw.Range("B4").value # 平均サービス率
    simtime = xw.Range("B5").value # シミュレーション時間

    # 到着間隔を生成する乱数発生器
    g1 = exponentialDistribution(1 / lmd)
    # サービス時間を生成する乱数発生器
    g2 = exponentialDistribution(1 / mu)
    # 窓口を示すファシリティ資源
    f = Facility(1, monitor = True)
    # 待ち時間を記録するモニター
    monitor = Monitor(["waitingTime"], ["f"], name = "wt")

    # 到着プロセス
    def arrive():
        while True:
            # 次の到着まで待つ
            yield pause(next(g1))
            # サービスプロセスを開始する
            # - この処理自体は非同期なので、一瞬で完了する
            yield subactivate(service)()

    # サービスプロセス
    def service():
        t0 = now()
        # 窓口の待ち行列に並ぶ
        # - 窓口を示すファシリティ資源の利用を予約する
        # - 既に利用者がいる場合は、利用可能になるまで待ち受ける
        lock = (yield f.request(name = "lock"))["lock"]
        # 窓口サービスを開始する
        # - 窓口を示すファシリティ資源の利用権を取得した
        # - 利用権を開放するまで他の利用者は利用できない
        t1 = now()
        # 待ち時間を記録
        monitor.observe(t1 - t0)
        # 窓口を利用する
        yield pause(next(g2))
        # 窓口を示すファシリティ資源の利用権を開放する
        lock.release()

    # 到着プロセスの開始を登録する
    activate(arrive)()

    # シミュレーション開始
    start(until = simtime)

    # M/M/1 の理論値
    rho = lmd / mu # 平均利用率
    Q = rho / (1.0 - rho) # 平均待ち行列
    W = Q / mu # 平均待ち時間

    waitingTime = monitor["waitingTime"].mean()
    # 平均待ち行列(窓口利用者をカウントしない)
    queue0 = f.monitor["要求待ち"].mean()
    # 平均システム内客数(窓口利用者をカウントする)
    queue = (f.monitor["要求待ち"] + f.monitor["バッファ"]).mean()
    #print(f"平均待ち時間: {waitingTime:.3f} (理論値: {W:.3f})")
    #print(f"平均システム内客数: {queue:.3f} (理論値: {Q:.3f})")
    xw.Range("B9").value = waitingTime
    xw.Range("B10").value = queue
    xw.Range("D9").value = W
    xw.Range("D10").value = Q

def clear():
    xw.Range("B9").value = ""
    xw.Range("B10").value = ""
    xw.Range("D9").value = ""
    xw.Range("D10").value = ""
